"""Definitions and behavior for iCalendar, also known as vCalendar 2.0"""

import behavior
import dateutil.rrule
import dateutil.tz
import parsing
import registry
import base64
import serializing
import StringIO
import datetime
import socket, random #for generating a UID
import itertools

from vobject import VObjectError, NativeError, ValidateError, VBase, Component,\
                    ContentLine

registry.registerTzid("UTC", dateutil.tz.tzutc())

#------------------------------- Constants -------------------------------------
DATENAMES=("rdate", "exdate")
RULENAMES=("exrule", "rrule")
PRODID = u"-//PYVOBJECT//NONSGML Version 1//EN"

#--------------------------- Helper functions ----------------------------------
def newrule(oldrule, newdt):
    """Take an rrule, return a new rrule with dtstart changed."""
    fields=['freq', 'interval', 'wkst', 'count', 'until', 'bysetpos',
            'bymonth', 'bymonthday', 'byyearday', 'byeaster',
            'byweekno', 'byweekday'] #, 'byhour', 'byminute', 'bysecond']
    newdict={'dtstart' : newdt}
    for name in fields: newdict[name] = getattr(oldrule, '_'+name)
    return dateutil.rrule.rrule(**newdict)

def backslashEscape(s):
    return s.replace("\\","\\\\").replace(";","\;").replace(",","\,").replace("\n","\\n")

#-------------------- Helper subclasses ----------------------------------------

class TimezoneComponent(Component):
    """A VTIMEZONE object.
    
    VTIMEZONEs are parsed by dateutil.tz.tzical, the resulting datetime.tzinfo
    subclass is stored in self.tzinfo, self.tzid stores the TZID associated
    with this timezone.
    
    @ivar name:
        The uppercased name of the object, in this case always 'VTIMEZONE'.
    @ivar tzinfo:
        A datetime.tzinfo subclass representing this timezone.
    @ivar tzid:
        The string used to refer to this timezone.
    
    """

    def __init__(self, component=None, tzinfo=None, *args, **kwds):
        """Accept an existing Component or a tzinfo class."""
        super(TimezoneComponent, self).__init__(*args, **kwds)
        self.name="VTIMEZONE"
        self.setTzinfo(None)
        self.isNative=True
        self.setTzinfo(tzinfo)
        if component is not None:
            self.behavior=component.behavior
            self.parentBehavior=component.parentBehavior
            tzstring=component.serialize()
            #rrule doesn't like unicode
            buffer=StringIO.StringIO(str(tzstring))
            buffer.seek(0)
            self.setTzinfo(dateutil.tz.tzical(buffer).get())

    def __setattr__(self, name, value):
        if name == 'tzinfo':
            self.setTzinfo(value)
        else:
            super(TimezoneComponent, self).__setattr__(name, value)

    def setTzinfo(self, tzinfo):
        """Set self.tzinfo and self.tzid, also register tzinfo."""
        object.__setattr__(self, 'tzinfo', tzinfo)
        tzid = self.pickTzid(tzinfo)
        if tzid: self.add(ContentLine('TZID', [], tzid))
        if tzid and not registry.getTzid(tzid):
            registry.registerTzid(tzid, tzinfo)

    def pickTzid(cls, tzinfo):
        """Given a tzinfo class, use the obvious TZID or pick a TZID."""
        if tzinfo is None: return None
        #If tzinfo is UTC, we don't need a TZID
        if tzinfo == registry.getTzid('UTC'): return None
        #For now, assume a tzical class
        return tzinfo._tzid
    pickTzid = classmethod(pickTzid)

    def __str__(self):
        return "<VTIMEZONE (tzinfo=" + str(self.tzinfo) +">"
    
    def __repr__(self):
        return self.__str__()
    
    def prettyPrint(self, level, tabwidth):
        pre = ' ' * level * tabwidth
        print pre, self.name
        print pre, "TZID: " + self.tzid[0]
        print


class RecurringComponent(Component):
    """A vCalendar component like VEVENT or VTODO which may recur.
        
    Any recurring component can have one or multiple RRULE, RDATE,
    EXRULE, or EXDATE lines, and one or zero DTSTART lines.  It can also have a
    variety of children that don't have any recurrence information.  Recurrence
    data is stored as a dateutil.rrule.rruleset in self.rruleset.
    
    Unfortunately, dateutil doesn't mimic rfc2445 in every way.  No distinction
    is made between dates and datetimes, so exclusion dates or rules that use
    dates instead of datetimes won't be processed correctly by rruleset.  Also,
    rrule doesn't know anything about duration of events.  For constant duration
    recurring events, this doesn't matter.  However, RDATEs can be specified
    with varying PERIOD, any RDATES which are parsed that have a PERIOD
    parameter will be added to self.rruleset AND stored as an array of
    ContentLines in self.clobberedRDates.  This array will be included when
    the RecurringComponent is serialized.
    
    Note that rruleset allows different rules to have different starting dates,
    but iCalendar components do not.  If you try to serialize an rruleset with
    different DTSTART dates, it will be serialized as a collection of RDATEs.
    At first glance it would seem that collections of RRULEs with different
    DTSTARTs could be implemented as an RRULE + an EXRULE, but EXRULEs take
    precedence over any RDATE or RRULE, so the resulting rule wouldn't be
    correct.
    
    Finally, in the example below, note that dtstart is included in the
    rruleset.  This is not the default behavior for dateutil's rrule 
    implementation unless dtstart would already have been a member of the 
    recurrence rule, and as a result, COUNT is wrong. This is worked around when
    parsing by adjusting count down by one if an rrule has a count and dtstart
    isn't in its result set, the opposite occurs when serializing.  But when
    creating rrule's programmatically it should be kept in mind that count
    doesn't necessarily mean what rfc2445 says.
    
    >>> import dateutil.rrule, datetime
    >>> vevent = RecurringComponent(name='VEVENT')
    >>> line = "RRULE:FREQ=WEEKLY;COUNT=1;INTERVAL=2;BYDAY=TU,TH"
    >>> vevent.rrule = [dateutil.rrule.rrulestr(line)]
    >>> vevent.setDtstart(datetime.datetime(2005, 1, 19, 9))
    >>> list(vevent.rruleset)
    [datetime.datetime(2005, 1, 19, 9, 0), datetime.datetime(2005, 1, 20, 9, 0)]

    Note that the properties associated with recurrence are accessed more
    directly than their counterparts in normal Components.  dtstart is just a
    datetime object, whereas dtend and dtstamp are lists containing one
    ContentLine whose value is a datetime object.  This 
    
    @ivar rruleset:
        A U{rruleset<https://moin.conectiva.com.br/DateUtil>}.
    @ivar dtstart:
        The first occurrence of the datetime set.
        
        dateutil.rrule.rrule doesn't automatically include dtstart in its
        result set, but iCalendar RRULEs must.  When a new RecurringComponent is
        created, and any time dtstart is changed, an RDATE for dtstart is added.
        If dtstart is changed, the new RDATE will replace any existing RDATE
        corresponding to the previous dtstart.  Any RDATE identical to dtstart
        will not be serialized.
    @ivar rrule:
        Getting self.rrule or self.exrule will return the relevant list of
        dateutil.rrule.rrule instances.  Setting replaces self.rruleset's list
        of rules.
    @ivar exrule:
        See L{rrule}.
    @ivar rdate:
        Getting self.rdate or self.exdate will return the relevant list of
        datetimes, setting a list of datetimes replaces or appends the
        exact datetimes to self.rruleset's lists.  Setting a list of dates (not
        datetimes) will set datetimes with time and timezone determined by
        self.dtstart.
    @ivar exdate:
        See L{rdate}.
    @ivar clobberedRDates:
        A list of RDATE ContentLines that represent a PERIOD of time.  dateutil
        can't work with these, so they're dealt with separately.
        
    """
    def __init__(self, component=None, *args, **kwds):
        """Take an existing Component, copy __dict__, create self.rruleset."""
        if not component:
            self.rruleset=dateutil.rrule.rruleset()
            super(RecurringComponent, self).__init__(*args, **kwds)
            self.isNative=True
            self.clobberedRDates=[]
            return
        else:
            self.__dict__ = component.__dict__.copy()
            
        #don't do anything more if component was already a RecurringComponent
        if isinstance(component, RecurringComponent): return True
        
        self.isNative=True
        self.clobberedRDates=[]
        self.rruleset=dateutil.rrule.rruleset()
        # Odd things will happen if there's no dtstart and there is recurrence
        # data.  TODO: These need more thought.
        try:
            component.dtstart[0].transformToNative()
            dtstart = component.dtstart[0].value
            self.setDtstart(dtstart)
        except AttributeError:
            dtstart = None
        for name in DATENAMES + RULENAMES:
            addfunc=getattr(self.rruleset, name)
            parsefunc=parsing.headerToParser[name]
            self.contents[name]=[]
            for line in component.contents.get(name, ()):
                if name in DATENAMES:
                    valueParam = line.params.get("VALUE", ["DATE-TIME"])[0]
                    if valueParam == "DATE-TIME":
                        map(addfunc, parsefunc(line))
                    elif valueParam == "DATE":
                        for dt in parsefunc(line):
                            addfunc(dtstart.replace(dt.year, dt.month, dt.day))
                    elif valueParam == "PERIOD" and name == "rdate":
                        # We lose information with RDATEs that have explicit
                        # duration, stick them in self.
                        self.clobberedRDates.append(line)
                        for dt, delta in parsefunc(line):
                            addfunc(dt)
                    else:
                        #todo: allow X- parameters
                        raise VObjectError
                elif name in RULENAMES:
                    addfunc(parsefunc(line, dtstart))
                    if name == 'rrule' and self.rrule[-1]._count != None:
                        #todo: it's conceivable that an rrule might have 0
                        #resulting datetimes, we should catch that exception
                        if self.rrule[-1][0] != dtstart:
                            self.rrule[-1]._count -= 1

    def __getattr__(self, name):
        """For convenience, make self.contents directly accessible."""
        if name in DATENAMES + RULENAMES:
            return getattr(self.rruleset, '_' + name)
        else:
            return super(RecurringComponent, self).__getattr__(name)

    def __setattr__(self, name, value):
        """For convenience, make self.contents directly accessible."""
        if name in RULENAMES+DATENAMES:
            container=getattr(self.rruleset, '_' + name)
        if name in DATENAMES:
            if type(value) != list:
                raise VObjectError("RecurringComponent children must be lists")
            for dt in value:
                if isinstance(dt, datetime.datetime):
                    container.append(dt)
                elif isinstance(dt, datetime.date):
                    newdt=self.dtstart[0].value.replace(dt.year,dt.month,dt.day)
                    container.append(newdt)
                else:
                    m="Recurrence dates can only be set to date or datetime"
                    raise VObjectError(m)
        elif name in RULENAMES:
            if type(value) != list:
                raise VObjectError("RecurringComponent children must be lists")

            for rule in value:
                if isinstance(rule, dateutil.rrule.rrule):
                    container.append(rule)
                else:
                    m="Recurrence rules must be dateutil.rrule.rrule instances"
                    raise VObjectError(m)
        elif name == 'rruleset' or name[0] == '_':
            object.__setattr__(self, name, value)
        else:
            super(RecurringComponent, self).__setattr__(name, value)

    def getDtstart(self):
        """Return dtstart if it's set, or None."""
        line = self.contents.get('dtstart', [None])[0]
        if line: return line.value
        else: return None

    def setDtstart(self, dt):
        """Set or reset dtstart for rrules and exrules. Change artificial RDATE.
        
        dateutil.rrule.rrule defaults to using now() as its dtstart.  It's
        convenient to be able to create rules with incorrect dtstart then set
        dtstart all at once.
        
        Deal with the RDATE hack to make rrule's contain dtstart. Note that
        other than this, rdate and exdate times aren't changed, so if they were
        set with DATE data, they'll be incorrect after setDtstart.
        
        """
        for key in RULENAMES:
            container=getattr(self.rruleset, '_' + key)
            newlist = [newrule(rule, dt) for rule in container]
            setattr(self.rruleset, '_' + key, newlist)
        if getattr(self, 'dtstart', None):
            olddt=self.getDtstart()
            self.dtstart=[]
            for i in xrange(len(self.rdate)):
                if self.rdate[i] == olddt: del self.rdate[i]
        self.add(ContentLine('DTSTART', [], dt, isNative=True))
        self.rdate.insert(0, dt)

    def getTzidIfNeeded(self):
        """Return dtstart's tzid if any rrule's exist, or None."""
        dt = self.getDtstart()
        if len(self.rrule) > 0 and dt and type(dt) == datetime.datetime and \
           dt.tzinfo <> registry.getTzid("UTC"):
                return TimezoneComponent.pickTzid(dt.tzinfo)
        return None

    def __str__(self):
        if self.name:
            return "<" + self.name + " (with rruleset)| " + str(self.getSortedChildren()) + ">"
        else:
            return '<' + '*unnamed*' + ' (with rruleset)| ' + str(self.getSortedChildren()) + '>'


class RecurringBehavior(behavior.Behavior):
    """Parent Behavior for components which should be RecurringComponents."""
    hasNative = True
    isComponent = True

    def transformToNative(cls, obj):
        """Turn a recurring Component into a RecurringComponent."""
        if obj.isNative: return obj
        else: return RecurringComponent(obj)

    def transformFromNative(cls, obj):
        """As a very rough initial cut, turn dates in rruleset into RDATEs.
        
        This obviously doesn't work for infinite RRULEs, arbitrarily we take the
        first 500 dates.
        
        """
        new = Component(obj.name)
        new.parentBehavior = obj.parentBehavior
        new.behavior = obj.behavior
        new.contents = obj.contents.copy()
        dtstart = obj.getDtstart()
        if not getattr(new, 'rdate', None): new.rdate=[]
        new.rdate.extend(obj.clobberedRDates)
        for dt in itertools.islice(obj.rruleset, 500):
            if dt == dtstart: continue
            new.add(ContentLine('RDATE', [], dt, isNative=True))
        return new
            
    def generateImplicitParameters(cls, obj):
        """Generate a UID if one does not exist.
        
        This is just a dummy implementation, for now.
        
        """
        if len(getattr(obj, 'uid', [])) == 0:
            rand = str(int(random.random() * 100000))
            now = datetime.datetime.now(registry.getTzid('UTC'))
            now = serializing.dateTimeToString(now)
            host = socket.gethostname()
            obj.add(ContentLine('UID', [], now + '-' + rand + '@' + host))        

    transformToNative = classmethod(transformToNative)
    transformFromNative = classmethod(transformFromNative)
    generateImplicitParameters = classmethod(generateImplicitParameters)
    
class DateTimeBehavior(behavior.Behavior):
    """Parent Behavior for ContentLines containing one DATE-TIME."""
    hasNative = True

    def transformToNative(cls, obj):
        """Turn obj.value into a datetime.

        RFC2445 allows times without time zone information, "floating times"
        in some properties.  Mostly, this isn't what you want, but when parsing
        a file, real floating times are noted by setting to 'TRUE' the
        X-VOBJ-FLOATINGTIME-ALLOWED parameter.
        
        If a TZID exists, the X-VOBJ-PRESERVE-TZID parameter will be set to
        'TRUE' so the TZID will be recreated when output.

        """
        if obj.isNative: return obj
        obj.isNative = True
        obj.value=str(obj.value)
        #we're cheating a little here, parseDtstart allows DATE
        obj.value=parsing.parseDtstart(obj)
        if not obj.value.tzinfo:
            obj.params['X-VOBJ-FLOATINGTIME-ALLOWED'] = ['TRUE']
        if obj.params.get('TZID'):
            obj.params['X-VOBJ-PRESERVE-TZID'] = ['TRUE']
            del obj.params['TZID']
        return obj

    def transformFromNative(cls, obj):
        """Replace the datetime in obj.value with an ISO 8601 string.
        
        Generally, it's best to express datetimes in UTC, so datetimes are
        converted to UTC unless a X-VOBJ-PRESERVE-TZID parameter is set to
        'TRUE'.

        While RFC2445 allows "floating times", it's highly unlikely you actually
        want one, so datetimes without tzinfo and without TZID are interpreted
        as being in the local timezone unless a X-VOBJ-FLOATINGTIME-ALLOWED
        parameter is present and 'TRUE'.

        """
        obj.isNative = False
        if obj.value.tzinfo is None:
            if obj.params.get('X-VOBJ-FLOATINGTIME-ALLOWED', [''])[0] <> 'TRUE':
                obj.value = obj.value.replace(tzinfo=dateutil.tz.tzlocal())
        preserveTZ = obj.params.get('X-VOBJ-PRESERVE-TZID', [''])[0] == 'TRUE'
        tzid = TimezoneComponent.pickTzid(obj.value.tzinfo)
        obj.value = serializing.dateTimeToString(obj.value, preserveTZ)
        if preserveTZ and tzid:
            obj.params['TZID'] = [tzid]
        if obj.params.get('X-VOBJ-FLOATINGTIME-ALLOWED'):
            del obj.params['X-VOBJ-FLOATINGTIME-ALLOWED']
        if obj.params.get('X-VOBJ-PRESERVE-TZID'):
            del obj.params['X-VOBJ-PRESERVE-TZID']

        return obj
    
    transformToNative = classmethod(transformToNative)
    transformFromNative = classmethod(transformFromNative)

class DateOrDateTimeBehavior(behavior.Behavior):
    """Parent Behavior for ContentLines containing one DATE or DATE-TIME."""
    hasNative = True

    def transformToNative(cls, obj):
        """Turn obj.value into a date or datetime."""
        if obj.isNative: return obj
        obj.isNative = True
        obj.value=str(obj.value)
        obj.value=parsing.parseDtstart(obj)
        if obj.params.get("VALUE", ["DATE-TIME"])[0] == 'DATE-TIME':
            if not obj.value.tzinfo:
                obj.params['X-VOBJ-FLOATINGTIME-ALLOWED'] = ['TRUE']
            if obj.params.has_key('TZID'): del obj.params['TZID']
        return obj

    def transformFromNative(cls, obj):
        """Replace the date or datetime in obj.value with an ISO 8601 string."""
        obj.isNative = False
        if type(obj.value) == datetime.date:
            obj.params['VALUE']=['DATE']
            obj.value = serializing.dateToString(obj.value)
            return obj
        else: return DateTimeBehavior.transformFromNative(obj)
    
    transformToNative = classmethod(transformToNative)
    transformFromNative = classmethod(transformFromNative)


class TextBehavior(behavior.Behavior):
    """Provide backslash escape encoding/decoding for single valued properties.
    
    TextBehavior also deals with base64 encoding if the ENCODING parameter is
    explicitly set to BASE64.
    
    """
    def decode(cls, line):
        """Remove backslash escaping from line.value."""
        if line.encoded:
            encoding = line.params.get('ENCODING')
            if encoding and encoding[0] == 'BASE64':
                line.value = base64.decodestring(line.value)
            else:
                line.value = parsing.stringToTextValues(line.value)[0]
            line.encoded=0
    
    def encode(cls, line):
        """Backslash escape line.value."""
        if not line.encoded:
            encoding = line.params.get('ENCODING')
            if encoding and encoding[0] == 'BASE64':
                line.value = base64.encodestring(line.value)
            else:
                line.value = backslashEscape(line.value)
            line.encoded=1

    decode       = classmethod(decode)
    encode       = classmethod(encode)


class MultiTextBehavior(behavior.Behavior):
    """Provide backslash escape encoding/decoding of each of several values.
    
    After transformation, value is a list of strings.
    
    """
    def decode(cls, line):
        """Remove backslash escaping from line.value, then split on commas."""
        if line.encoded:
            line.value = parsing.stringToTextValues(line.value)
            line.encoded=0
    
    def encode(cls, line):
        """Backslash escape line.value."""
        if not line.encoded:
            line.value = map(backslashEscape, line.value).join(',')
            line.encoded=1
            
    decode       = classmethod(decode)
    encode       = classmethod(encode)
    

#------------------------ Registered Behavior subclasses -----------------------
class VCalendar2_0(behavior.Behavior):
    """vCalendar 2.0 behavior."""
    name = 'VCALENDAR'
    description = 'vCalendar 2.0, also known as iCalendar.'
    versionString = '2.0'
    isComponent = True
    knownChildren = {'CALSCALE':  (0, 1, None),#min, max, behaviorRegistry id
                     'METHOD':    (0, 1, None),
                     'VERSION':   (1, 1, None),
                     'PRODID':    (1, 1, None),
                     'VTIMEZONE': (0, None, None),
                     'VEVENT':    (0, None, None),
                     'VTODO':     (0, None, None),
                     'VJOURNAL':  (0, None, None),
                     'VFREEBUSY': (0, None, None)
                    }
                    
    def generateImplicitParameters(cls, obj):
        """Create PRODID, VERSION, and VTIMEZONEs if needed.
        
        Generally, it's best to express datetimes in UTC.  But RFC2445 Page 117:
                
           When used with a recurrence rule, the "DTSTART" and "DTEND"
           properties MUST be specified in local time and the appropriate
           set of "VTIMEZONE" calendar components MUST be included. For
           detail on the usage of the "VTIMEZONE" calendar component, see
           the "VTIMEZONE" calendar component definition.
        
        VTIMEZONEs will need to exist whenever TZID parameters exist, when
        native datetimes have a X-VOBJ-PRESERVE-TZID parameter set, or when a
        RecurringComponent children's getTzidIfNeeded method returns a TZID.
        
        """
        if len(getattr(obj, 'prodid', [])) == 0:
            obj.add(ContentLine('PRODID', [], PRODID))
        if len(getattr(obj, 'version', [])) == 0:
            obj.add(ContentLine('VERSION', [], '2.0'))
        tzidsUsed = {}
        
        def findTzids(obj, table):
            if isinstance(obj, ContentLine):
                if obj.params.get('TZID'):
                    table[obj.params.get('TZID')[0]]=1
                elif obj.params.get('X-VOBJ-PRESERVE-TZID'):
                    tzinfo = getattr(obj.value, 'tzinfo')
                    tzid = TimezoneComponent.pickTzid(tzinfo)
                    if tzid: table[tzid]=1
            if isinstance(obj, RecurringComponent):
                tzid = obj.getTzidIfNeeded()
                if tzid: table[tzid]=1
            for child in obj.getChildren():
                if obj.name is not 'VTIMEZONE': findTzids(child, table)
        
        findTzids(obj, tzidsUsed)
        oldtzids = [x.tzid[0] for x in getattr(obj, 'vtimezone', [])]
        for tzid in tzidsUsed.keys():
            if oldtzids.index(tzid) >= 0: continue
            obj.add(TimezoneComponent(tzinfo=registry.getTzid(tzid)))
    
    generateImplicitParameters = classmethod(generateImplicitParameters)
registry.registerBehavior(VCalendar2_0, default=True)

class VTimezone(behavior.Behavior):
    """Timezone behavior."""
    name = 'VTIMEZONE'
    hasNative = True
    isComponent = True
    description = 'A grouping of component properties that defines a time zone.'
    knownChildren = {'TZID':         (1, 1, None),#min, max, behaviorRegistry id
                     'LAST-MODIFIED':(0, 1, None),
                     'TZURL':        (0, 1, None),
                     'STANDARD':     (0, None, None),#NOTE: One of Standard or
                     'DAYLIGHT':     (0, None, None) #      Daylight must appear
                    }

    def validate(cls, obj, raiseException, *args):
        return True #TODO: FIXME
        if obj.contents.has_key('standard') or obj.contents.has_key('daylight'):
            return super(VTimezone, cls).validate(obj, raiseException, *args)
        else:
            if raiseException:
                m = "VTIMEZONE components must contain a STANDARD or a DAYLIGHT\
                     component"
                raise ValidateError(m)
            return False

    def transformToNative(cls, obj):
        """Turn a recurring Component into a RecurringComponent."""
        if obj.isNative: return obj
        return TimezoneComponent(obj)

    def transformFromNative(cls, obj):
        """Calculate ContentLines for rruleset and dtstart."""
        #TODO
        raise NativeError("No transformFromNative defined")        
        
    validate = classmethod(validate)    
    transformToNative = classmethod(transformToNative)
    transformFromNative = classmethod(transformFromNative)
registry.registerBehavior(VTimezone)

class VEvent(RecurringBehavior):
    """Event behavior."""
    name='VEVENT'
    description='A grouping of component properties, and possibly including \
                 "VALARM" calendar components, that represents a scheduled \
                 amount of time on a calendar.'
    knownChildren = {'DTSTART':      (0, 1, None),#min, max, behaviorRegistry id
                     'CLASS':        (0, 1, None),  
                     'CREATED':      (0, 1, None),
                     'DESCRIPTION':  (0, 1, None),  
                     'GEO':          (0, 1, None),  
                     'LAST-MODIFIED':(0, 1, None),
                     'LOCATION':     (0, 1, None),  
                     'ORGANIZER':    (0, 1, None),  
                     'PRIORITY':     (0, 1, None),  
                     'DTSTAMP':      (0, 1, None),
                     'SEQUENCE':     (0, 1, None),  
                     'STATUS':       (0, 1, None),  
                     'SUMMARY':      (0, 1, None),                     
                     'TRANSP':       (0, 1, None),  
                     'UID':          (0, 1, None),  
                     'URL':          (0, 1, None),  
                     'RECURRENCE-ID':(0, 1, None),  
                     'DTEND':        (0, 1, None), #NOTE: Only one of DtEnd or
                     'DURATION':     (0, 1, None), #      Duration can appear
                     'ATTACH':       (0, None, None),
                     'ATTENDEE':     (0, None, None),
                     'CATEGORIES':   (0, None, None),
                     'COMMENT':      (0, None, None),
                     'CONTACT':      (0, None, None),
                     'EXDATE':       (0, None, None),
                     'EXRULE':       (0, None, None),
                     'REQUEST-STATUS': (0, None, None),
                     'RELATED-TO':   (0, None, None),
                     'RESOURCES':    (0, None, None),
                     'RDATE':        (0, None, None),
                     'RRULE':        (0, None, None),
                     'VALARM':       (0, None, None)
                    }

    def validate(cls, obj, raiseException, *args):
        if obj.contents.has_key('DTEND') and obj.contents.has_key('DURATION'):
            if raiseException:
                m = "VEVENT components cannot contain both DTEND and DURATION\
                     components"
                raise ValidateError(m)
            return False
        else:
            return super(VEvent, cls).validate(obj, raiseException, *args)
      
    validate = classmethod(validate)
registry.registerBehavior(VEvent)


class VTodo(RecurringBehavior):
    """To-do behavior."""
    name='VTODO'
    description='A grouping of component properties and possibly "VALARM" \
                 calendar components that represent an action-item or \
                 assignment.'
    knownChildren = {'DTSTART':      (0, 1, None),#min, max, behaviorRegistry id
                     'CLASS':        (0, 1, None),
                     'COMPLETED':    (0, 1, None),
                     'CREATED':      (0, 1, None),
                     'DESCRIPTION':  (0, 1, None),  
                     'GEO':          (0, 1, None),  
                     'LAST-MODIFIED':(0, 1, None),
                     'LOCATION':     (0, 1, None),  
                     'ORGANIZER':    (0, 1, None),  
                     'PERCENT':      (0, 1, None),  
                     'PRIORITY':     (0, 1, None),  
                     'DTSTAMP':      (0, 1, None),
                     'SEQUENCE':     (0, 1, None),  
                     'STATUS':       (0, 1, None),  
                     'SUMMARY':      (0, 1, None),                     
                     'UID':          (0, 1, None),  
                     'URL':          (0, 1, None),  
                     'RECURRENCE-ID':(0, 1, None),  
                     'DUE':          (0, 1, None), #NOTE: Only one of Due or
                     'DURATION':     (0, 1, None), #      Duration can appear
                     'ATTACH':       (0, None, None),
                     'ATTENDEE':     (0, None, None),
                     'CATEGORIES':   (0, None, None),
                     'COMMENT':      (0, None, None),
                     'CONTACT':      (0, None, None),
                     'EXDATE':       (0, None, None),
                     'EXRULE':       (0, None, None),
                     'REQUEST-STATUS': (0, None, None),
                     'RELATED-TO':   (0, None, None),
                     'RESOURCES':    (0, None, None),
                     'RDATE':        (0, None, None),
                     'RRULE':        (0, None, None),
                     'VALARM':       (0, None, None)
                    }

    def validate(cls, obj, raiseException, *args):
        if obj.contents.has_key('DUE') and obj.contents.has_key('DURATION'):
            if raiseException:
                m = "VEVENT components cannot contain both DTEND and DURATION\
                     components"
                raise ValidateError(m)
            return False
        else:
            return super(VTodo, cls).validate(obj, raiseException, *args)
      
    validate = classmethod(validate)
registry.registerBehavior(VTodo)


class VJournal(RecurringBehavior):
    """Journal entry behavior."""
    name='VJOURNAL'
    knownChildren = {'DTSTART':      (0, 1, None),#min, max, behaviorRegistry id
                     'CLASS':        (0, 1, None),  
                     'CREATED':      (0, 1, None),
                     'DESCRIPTION':  (0, 1, None),  
                     'LAST-MODIFIED':(0, 1, None),
                     'ORGANIZER':    (0, 1, None),  
                     'DTSTAMP':      (0, 1, None),
                     'SEQUENCE':     (0, 1, None),  
                     'STATUS':       (0, 1, None),  
                     'SUMMARY':      (0, 1, None),                     
                     'UID':          (0, 1, None),  
                     'URL':          (0, 1, None),  
                     'RECURRENCE-ID':(0, 1, None),  
                     'ATTACH':       (0, None, None),
                     'ATTENDEE':     (0, None, None),
                     'CATEGORIES':   (0, None, None),
                     'COMMENT':      (0, None, None),
                     'CONTACT':      (0, None, None),
                     'EXDATE':       (0, None, None),
                     'EXRULE':       (0, None, None),
                     'REQUEST-STATUS': (0, None, None),
                     'RELATED-TO':   (0, None, None),
                     'RDATE':        (0, None, None),
                     'RRULE':        (0, None, None)
                    }
registry.registerBehavior(VJournal)


class VFreeBusy(behavior.Behavior):
    """Free/busy state behavior."""
    name='VFREEBUSY'
    description='A grouping of component properties that describe either a \
                 request for free/busy time, describe a response to a request \
                 for free/busy time or describe a published set of busy time.'
    knownChildren = {'DTSTART':      (0, 1, None),#min, max, behaviorRegistry id
                     'CONTACT':      (0, 1, None),
                     'DTEND':        (0, 1, None),
                     'DURATION':     (0, 1, None),
                     'ORGANIZER':    (0, 1, None),  
                     'DTSTAMP':      (0, 1, None), 
                     'UID':          (0, 1, None),  
                     'URL':          (0, 1, None),   
                     'ATTENDEE':     (0, None, None),
                     'COMMENT':      (0, None, None),
                     'FREEBUSY':     (0, None, None),
                     'REQUEST-STATUS': (0, None, None)
                    }
registry.registerBehavior(VFreeBusy)


class VAlarm(behavior.Behavior):
    """Alarm behavior."""
    name='VALARM'
    isComponent = True
    description='Alarms describe when and how to provide alerts about events \
                 and to-dos.'
    knownChildren = {'ACTION':       (1, 1, None),#min, max, behaviorRegistry id
                     'TRIGGER':      (1, 1, None),  
                     'DURATION':     (0, 1, None),
                     'REPEAT':       (0, 1, None)
                    }

    def validate(cls, obj, raiseException, *args):
        """
        #TODO
     audioprop  = 2*(

                ; 'action' and 'trigger' are both REQUIRED,
                ; but MUST NOT occur more than once

                action / trigger /

                ; 'duration' and 'repeat' are both optional,
                ; and MUST NOT occur more than once each,
                ; but if one occurs, so MUST the other

                duration / repeat /

                ; the following is optional,
                ; but MUST NOT occur more than once

                attach /

     dispprop   = 3*(

                ; the following are all REQUIRED,
                ; but MUST NOT occur more than once

                action / description / trigger /

                ; 'duration' and 'repeat' are both optional,
                ; and MUST NOT occur more than once each,
                ; but if one occurs, so MUST the other

                duration / repeat /

     emailprop  = 5*(

                ; the following are all REQUIRED,
                ; but MUST NOT occur more than once

                action / description / trigger / summary

                ; the following is REQUIRED,
                ; and MAY occur more than once

                attendee /

                ; 'duration' and 'repeat' are both optional,
                ; and MUST NOT occur more than once each,
                ; but if one occurs, so MUST the other

                duration / repeat /

     procprop   = 3*(

                ; the following are all REQUIRED,
                ; but MUST NOT occur more than once

                action / attach / trigger /

                ; 'duration' and 'repeat' are both optional,
                ; and MUST NOT occur more than once each,
                ; but if one occurs, so MUST the other

                duration / repeat /

                ; 'description' is optional,
                ; and MUST NOT occur more than once

                description /
        if obj.contents.has_key('DTEND') and obj.contents.has_key('DURATION'):
            if raiseException:
                m = "VEVENT components cannot contain both DTEND and DURATION\
                     components"
                raise ValidateError(m)
            return False
        else:
            return super(VEvent, cls).validate(obj, raiseException, *args)
        """
        return True
    
    validate = classmethod(validate)
registry.registerBehavior(VAlarm)

class Trigger(behavior.Behavior):
    """DATE-TIME or DURATION"""
    name='TRIGGER'
    pass#TODO

class FreeBusy(behavior.Behavior):
    """Free or busy period of time."""
    name='FREEBUSY'
    pass#TODO


#------------------------ Registration of common classes -----------------------

dateTimeList = ['LAST-MODIFIED', 'CREATED', 'COMPLETED', 'DTSTAMP']
map(lambda x: registry.registerBehavior(DateTimeBehavior, x), dateTimeList)

dateTimeOrDateList = ['DTEND', 'DTSTART', 'DUE', 'RECURRENCE-ID', 'RDATE']
map(lambda x: registry.registerBehavior(DateOrDateTimeBehavior, x),
    dateTimeOrDateList)

textList = ['CALSCALE', 'METHOD', 'PRODID', 'CLASS', 'COMMENT', 'DESCRIPTION',
            'LOCATION', 'STATUS', 'SUMMARY', 'TRANSP', 'CONTACT', 'RELATED-TO',
            'UID', 'ACTION', 'REQUEST-STATUS', 'TZID']
map(lambda x: registry.registerBehavior(TextBehavior, x), textList)

multiTextList = ['CATEGORIES', 'RESOURCES']
map(lambda x: registry.registerBehavior(MultiTextBehavior, x), multiTextList)


#------------------- Testing and running functions -----------------------------
if __name__ == '__main__':
    import tests
    tests._test()