"""vobject module for reading vCard and vCalendar files."""

import logging
import registry
import StringIO
import string
import exceptions

#------------------------------------ Logging ----------------------------------
logger = logging.getLogger('vobject')
if not logger.handlers:
    handler=logging.StreamHandler()
    formatter = logging.Formatter('%(name)s %(levelname)s %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
#logger.setLevel(logging.ERROR) #Display errors
logger.setLevel(logging.FATAL) #Don't display errors

#----------------------------------- Constants ---------------------------------
CR     = unichr(13)
LF     = unichr(10)
CRLF   = CR + LF
SPACE  = unichr(32)
TAB    = unichr(9)

#-------------------------------- Useful modules -------------------------------
#   use doctest, it kills two birds with one stone and docstrings often become
#                more readable to boot (see parseLine's docstring).
#   use logging, then when debugging we can just set our verbosity.
#   use epydoc syntax for documenting code, please document every class and non-
#                trivial method (see http://epydoc.sourceforge.net/epytext.html
#                and http://epydoc.sourceforge.net/fields.html).  Also, please
#                follow http://www.python.org/peps/pep-0257.html for docstrings.
#-------------------------------------------------------------------------------

#--------------------------------- Main classes --------------------------------
class VBase(object):
    """Base class for ContentLine and Component.
    
    @ivar behavior:
        The Behavior class associated with this object, which controls
        validation, transformations, and encoding.
    @ivar parentBehavior:
        The object's parent's behavior, or None if no behaviored parent exists.
    @ivar isNative:
        Boolean describing whether this component is a Native instance.
    
    """
    def __init__(self, *args, **kwds):
        super(VBase, self).__init__(*args, **kwds)
        self.behavior   = None
        self.parentBehavior = None
        self.isNative = False
    
    def validate(self, *args, **kwds):
        """Call the behavior's validate method, or return True."""
        if self.behavior:
            return self.behavior.validate(self, *args, **kwds)
        else: return True

    def getChildren(self):
        """Return an iterable containing the contents of the object."""
        return []

    def clearBehavior(self, cascade=True):
        """Set behavior to None. Do for all descendants if cascading."""
        self.behavior=None
        if cascade:
            for obj in self.getChildren():
                obj.behavior=None
                obj.parentBehavior=None
                obj.clearBehavior(True)

    def autoBehavior(self, cascade=False):
        """Set behavior if name is in self.parentBehavior.knownChildren.
        
        If cascade is True, unset behavior and parentBehavior for all
        descendants, then recalculate behavior and parentBehavior.
        
        """
        self.clearBehavior(cascade)
        try:
            id=self.parentBehavior.knownChildren[self.name][2]
            self.setBehavior(registry.getBehavior(self.name, id), cascade)
            if isinstance(self, ContentLine) and self.encoded:
                self.behavior.decode(self)
        except (KeyError, AttributeError):
            pass

    def setBehavior(self, behavior, cascade=True):
        """Set behavior. If cascade is True, autoBehavior all descendants."""
        self.behavior=behavior
        if cascade:
            for obj in self.getChildren():
                obj.parentBehavior=behavior
                obj.autoBehavior(True)

    def transformToNative(self):
        """Transform this object into a custom VBase subclass.
        
        transformToNative should always return a representation of this object.
        It may do so by modifying self in place then returning self, or by
        creating a new object.
        
        """
        if self.isNative or not self.behavior or not self.behavior.hasNative:
            return self
        else: return self.behavior.transformToNative(self)

    def transformChildrenToNative(self):
        """Recursively replace children with their native representation."""
        pass

    def transformFromNative(self):
        """Return self transformed into a ContentLine or Component if needed.
        
        May have side effects.  If it does, transformFromNative and
        transformToNative MUST have perfectly inverse side effects. Allowing
        such side effects is convenient for objects whose transformations only
        change a few attributes.
        
        Note that it isn't always possible for transformFromNative to be a
        perfect inverse of transformToNative, in such cases transformFromNative
        should return a new object, not self after modifications.
        
        """
        if not self.behavior or not self.behavior.hasNative:
            return self
        else: return self.behavior.transformFromNative(self)

    def serialize(self, buf=None, lineLength=75, validate=True):
        """Serialize to buf if it exists, otherwise return a string.
        
        Use self.behavior.serialize if behavior exists.
        
        """
        undoTransform = False
        if self.behavior:
            self.behavior.generateImplicitParameters(self)
            if validate: self.behavior.validate(self, raiseException=True)
            if self.isNative:
                transformed = self.transformFromNative()
                undoTransform = True
            else:
                transformed = self
            out = self.behavior.serialize(transformed, buf, lineLength)
            if undoTransform: self.transformToNative()
            return out
        else:
            return defaultSerialize(self, buf, lineLength)


class ContentLine(VBase):
    """Holds one content line for formats like vCard and vCalendar.

    For example::
      <SUMMARY{u'param1' : [u'val1'], u'param2' : [u'val2']}Bastille Day Party>

    @ivar name:
        The uppercased name of the contentline.
    @ivar params:
        A dictionary of parameters and associated lists of values (the list may
        be empty for empty parameters).
    @ivar value:
        The value of the contentline.
    @ivar singletonparams:
        A list of paramaters for which it's unclear if the string represents the
        parameter name or the parameter value. In vCard 2.1, "The value string
        can be specified alone in those cases where the value is unambiguous".
        This is crazy, but we have to deal with it.
    @ivar corruption:
        If this data resulted in an error during parsing, this will contain
        info, such as a raw unparsed line.
    @note: corruption is currently unused.
    @ivar encoded:
        A boolean describing whether the data in the content line is encoded.
        Generally, text read from a serialized vCard or vCalendar should be
        considered encoded.  Data added programmatically should not be encoded.
                    
    """
    def __init__(self, name, params, value, encoded=False, corruption=None,
                            isNative=False, *args, **kwds):
        """Take output from parseLine, convert params list to dictionary."""
        super(ContentLine, self).__init__(*args, **kwds)
        self.name        = name.upper()
        self.value       = value
        self.corruption  = corruption
        self.encoded     = encoded
        self.params      = {}
        self.singletonparams = []
        self.isNative = isNative
        def updateTable(x):
            if len(x) == 1: self.singletonparams += x
            else: self.params[x[0].upper()] = x[1:] #assumes no repeated params
        map(updateTable, params)
        qp = False
        if 'ENCODING' in self.params:
            if 'QUOTED-PRINTABLE' in self.params['ENCODING']:
                qp = True
                self.params['ENCODING'].remove('QUOTED-PRINTABLE')
                if 0==len(self.params['ENCODING']):
                    del self.params['ENCODING']
        if 'QUOTED-PRINTABLE' in self.singletonparams:
            qp = True
            self.singletonparams.remove('QUOTED-PRINTABLE')
        if qp:
            self.value = str(self.value).decode('quoted-printable')

    def __eq__(self, other):
        return (self.name == other.name) and (self.params == other.params) and (self.value == other.value)

    def __str__(self):
        return "<" + str(self.name) + str(self.params) + str(self.value) + ">"

    def __repr__(self):
        return self.__str__()

    def prettyPrint(self, level = 0, tabwidth=3):
        pre = ' ' * level * tabwidth
        print pre, self.name
        if self.params:
            lineKeys= self.params.keys()
            print pre, "params for ", self.name +':'
            for aKey in lineKeys:
                print pre + ' ' * tabwidth, aKey, self.params[aKey]
        print

class Component(VBase):
    """A complex property that can contain multiple ContentLines.
    
    For our purposes, a component must start with a BEGIN:xxxx line and end with
    END:xxxx, or have a PROFILE:xxx line if a top-level component.

    @ivar contents:
        A dictionary of lists of Component or ContentLine instances. The keys
        are the lowercased names of child ContentLines or Components.
        Note that BEGIN and END ContentLines are not included in contents.
    @ivar name:
        Uppercase string used to represent this Component, i.e VCARD if the
        serialized object starts with BEGIN:VCARD.
    @ivar useBegin:
        A boolean flag determining whether BEGIN: and END: lines should
        be serialized.

    """
    def __init__(self, name=None, *args, **kwds):
        super(Component, self).__init__(*args, **kwds)
        self.contents  = {}
        if name:
            self.name=name.upper()
            self.useBegin = True
        else:
            self.name = ''
            self.useBegin = False
        
        self.autoBehavior()

    def setProfile(self, name):
        """Assign a PROFILE to this unnamed component.
        
        Used by vCard, not by vCalendar.
        
        """
        if self.name or self.useBegin:
            if self.name == name: return
            raise VObjectError("This component already has a PROFILE or uses BEGIN.")
        self.name = name

    def __getattr__(self, name):
        """For convenience, make self.contents directly accessible."""
        try:
            return self.contents[name]
        except KeyError:
            raise exceptions.AttributeError, name

    def __setattr__(self, name, value):
        """For convenience, make self.contents directly accessible.
        
        self.contents must contain lists, raise an error if value isn't a list.
        Wrap NativeWrapper around object lists that aren't VBase descendants. 
        
        """
        vars = ['contents', 'name', 'behavior', 'parentBehavior']
        if name not in vars and name.lower()==name:
            if type(value) == list:
                self.contents[name]=value
            else:
                raise VObjectError("Component children must be lists")
        else:
            object.__setattr__(self, name, value)

    def add(self, objOrName):
        """Add objOrName to contents, set behavior if it can be inferred.
        
        If objOrName is a string, create an empty component or line based on
        behavior. If no behavior is found for the object, add a ContentLine.
        
        """
        if isinstance(objOrName, VBase):
            obj = objOrName
            if self.behavior:
                obj.parentBehavior = self.behavior
                obj.autoBehavior(True)
        else:
            name = objOrName.upper()
            try:
                id=self.behavior.knownChildren[name][2]
                behavior = registry.getBehavior(name, id)
                if behavior.isComponent:
                    obj = Component(name)
                else:
                    obj = ContentLine(name, [], '')
                obj.parentBehavior = self.behavior
                obj.behavior = behavior
                obj = obj.transformToNative()     
            except (KeyError, AttributeError):
                obj = ContentLine(objOrName, [], '') 
        self.contents.setdefault(obj.name.lower(), []).append(obj)

    def getChildren(self):
        for objList in self.contents.values():
            for obj in objList: yield obj

    def getSortedChildren(self):
        keys=self.contents.keys()
        keys.sort()
        l=[]
        for key in keys:
            for obj in self.contents[key]: l.append(obj)
        return l

    def setBehaviorFromVersionLine(self, versionLine):
        """Set behavior if one matches name, versionLine.value."""
        v=registry.getBehavior(self.name, versionLine.value)
        if v: self.setBehavior(v)
        
    def transformChildrenToNative(self):
        """Recursively replace children with their native representation.
        
        Timezones are transformed first, because other transforms may need the
        timezones defined.
        
        """
        timezones=self.contents.get('vtimezone', ())
        for i in xrange(len(timezones)):
            timezones[i]=timezones[i].transformToNative()
            timezones[i].transformChildrenToNative()
        for childArray in self.contents.values():
            for i in xrange(len(childArray)):
                childArray[i]=childArray[i].transformToNative()
                childArray[i].transformChildrenToNative()
    
    def __str__(self):
        if self.name:
            return "<" + self.name + "| " + str(self.getSortedChildren()) + ">"
        else:
            return '<' + '*unnamed*' + '| ' + str(self.getSortedChildren()) + '>'

    def __repr__(self):
        return self.__str__()

    def prettyPrint(self, level = 0, tabwidth=3):
        pre = ' ' * level * tabwidth
        print pre, self.name
        if isinstance(self, Component):
            for line in self.getChildren():
                line.prettyPrint(level + 1, tabwidth)

class VObjectError(Exception):
    def __init__(self, message):
        self.message = message
    def __str__(self):
        return repr(self.message)

class ParseError(VObjectError):
    pass

class ValidateError(VObjectError):
    pass

class NativeError(VObjectError):
    pass

#-------------------------- Parsing functions ----------------------------------
testLines="""
Line 0 text
 , Line 0 continued.
Line 1;encoding=quoted-printable:this is an evil=
 evil=
 format.
Line 2 is a new line, it does not start with whitespace.
"""

def getLogicalLines(fp, allowQP=True):
    """Iterate through a stream, yielding one logical line at a time.

    Because many applications still use vCard 2.1, we have to deal with the
    quoted-printable encoding for long lines, as well as the vCard 3.0 and
    vCalendar line folding technique, a whitespace character at the start
    of the line.
       
    Quoted-printable data will be decoded in the Behavior decoding phase.
       
    >>> import StringIO
    >>> f=StringIO.StringIO(testLines)
    >>> for n, l in enumerate(getLogicalLines(f)):
    ...     print "Line %s: %s" % (n, l)
    ...
    Line 0: Line 0 text, Line 0 continued.
    Line 1: Line 1;encoding=quoted-printable:this is an evil=
     evil=
     format.
    Line 2: Line 2 is a new line, it does not start with whitespace.

    """
    quotedPrintable=False
    logicalLine=""
    for line in fp:
        line = line.rstrip(CRLF)
        if not line or not line.rstrip():
            logger.warning("RFC2445 seems to prohibit blank lines, but we skip them.")
            if logicalLine: yield logicalLine
            logicalLine=""
            quotedPrintable=False
            continue

        if quotedPrintable and allowQP:
            logicalLine += '\n' + line
            quotedPrintable=False
        elif line[0] in (SPACE, TAB):
            if not logicalLine:
                logger.warning("Logical line started with whitespace.") 
            logicalLine += line[1:]
        elif logicalLine:
            yield logicalLine
            logicalLine=line
        else:
            logicalLine=line
        
        # hack to deal with the fact that vCard 2.1 allows parameters to be
        # encoded without a parameter name.  False positives are unlikely, but
        # possible.
        if allowQP: 
            if logicalLine[-1]=='=' and logicalLine.lower().find('quoted-printable') >= 0:
                quotedPrintable=True

    if logicalLine: yield logicalLine


#@@ issues remaining:
# need to test error codepaths to ensure errorhandling actually works as expected
def parseLine(line, strict=False):
    """Break an unfolded content line into name, parameters, value.
    
    The current error philosophy is to try extracting as much useful
    information as possible, even in the face of errors. A possible
    improvement is signalling an error containing as much of the parse
    as possible, or some multiple return value strategy.

    >>> parseLine(":")
    ('', [], '')
    >>> parseLine("blah:")
    ('BLAH', [], '')
    >>> parseLine("RDATE:VALUE=DATE:19970304,19970504,19970704,19970904")
    ('RDATE', [], 'VALUE=DATE:19970304,19970504,19970704,19970904')
    >>> parseLine('DESCRIPTION;ALTREP="http://www.wiz.org":The Fall 98 Wild Wizards Conference - - Las Vegas, NV, USA')
    ('DESCRIPTION', [['ALTREP', 'http://www.wiz.org']], 'The Fall 98 Wild Wizards Conference - - Las Vegas, NV, USA')
    >>> parseLine("EMAIL;PREF;INTERNET:john@nowhere.com")
    ('EMAIL', [['PREF'], ['INTERNET']], 'john@nowhere.com')
    >>> parseLine('EMAIL;TYPE="blah",hah;INTERNET="DIGI",DERIDOO:john@nowhere.com')
    ('EMAIL', [['TYPE', 'blah', 'hah'], ['INTERNET', 'DIGI', 'DERIDOO']], 'john@nowhere.com')
    >>> parseLine("NOTVALID", False)
    ('NOTVALID', [], '')
    >>> parseLine("NOTVALID", True)
    Traceback (most recent call last):
        ...
    ParseError: 'got end-of-line while reading in line name: NOTVALID'
    
    """
    # helpers
    class Cargo:
        """Carry the intermediate processing variables for the state machine."""
        def __init__(self):
            self.singleParamList=[]
            self.params=[]
            self.scratch=""
            self.name=""
            self.value=""

    def finishScratch(cargo):
        """Add scratch to the current parameter, empty scratch."""
        cargo.singleParamList.append(cargo.scratch)
        cargo.scratch=""

    def finishParam(cargo):
        """Finish scratch, empty singleParamList after adding it to params."""
        finishScratch(cargo)
        cargo.params.append(cargo.singleParamList)
        cargo.singleParamList=[]

    def error(msg):
        if strict:
            raise ParseError(msg)
        else:
            logger.error(msg)

    def ianaChar(c):
        return (c in string.letters) or (c in string.digits) or (c == "-")

    #vars which control state machine
    charIterator = enumerate(line)
    state        = "getting name"
    cargo        = Cargo()

    while True:
        try:
            charIndex, char = charIterator.next()
        except:
            charIndex += 1
            char = "eof"

        logger.debug(char + " <- " + state)

        if state == "getting name":
            if char == ';':     state = "getting param-name"
            elif char == ':':   state = "getting value"
            elif char == "eof":
                state = "error"
                error("got end-of-line while reading in line name: " + line)
#             elif ianaChar(char):
#                 cargo.name += char.upper()
#             else: state = "error"
            else: cargo.name += char.upper()
            
        elif state == "getting param-name":
            if char == ';':                  #valueless param, not end of params
                state = "getting param-name"
                finishParam(cargo)
            elif char == ':':                #valueless param, end of params
                state = "getting value"
                finishParam(cargo)
            elif char == '=':                #param with value
                state = "getting param-value"
                finishScratch(cargo)
            elif char == "eof":
                state = "error"
                error("got end-of-line while reading in param-name: " + line)
                finishParam(cargo)
#             elif ianaChar(char):
#                 cargo.scratch += char.upper()
#             else: state = "error"
            else: cargo.scratch += char.upper()
            
        elif state == "getting param-value":
            if char == ',':
                state = "getting param-value"
                finishScratch(cargo)
            elif char == ';':
                state = "getting param-name"
                finishParam(cargo)
            elif char == ':':
                state = "getting value"
                finishParam(cargo)
            elif char == '"': state = "getting param-value (quoted)"
            elif char == "eof":
                state = "error"
                error("got end-of-line while reading in param-value: " + line)
                finishParam(cargo)
            else: cargo.scratch += char

        elif state == "getting param-value (quoted)":
            if char == '"': state = "getting param-value (quoted) (end)"
            elif char == "eof":
                state  = "error"
                error("got end-of-line while reading in param-value (quoted): " + line)
            else: cargo.scratch += char

        elif state == "getting param-value (quoted) (end)":
            if char == ';':
                state = "getting param-name"
                finishParam(cargo)
            elif char == ',':
                state = "getting param-value"
                finishScratch(cargo)
            elif char == ':':
                state = "getting value"
                finishParam(cargo)
            elif char == "eof":
                state = "error"
                error("got end-of-line while reading in param-value (quoted) (end): " + line)
                finishParam(cargo)
            else:
                state = "error"
                error("did not expect this char at the end of a quoted param: '%s' in %s" % (char, line))
                finishParam(cargo)

        elif state == "getting value":
            state = "end"
            cargo.value = line[charIndex:]
        
        elif state == "end":    #an end state
            return cargo.name, cargo.params, cargo.value

        elif state == "error":  #an end state
            #Return as much valid data as possible. An alternative is to
            #signal an error and have the caller decide what to do.
            #Unfortunately, Python does not support "warnings" up the
            #call stack.
            return cargo.name, cargo.params, cargo.value

        else:
            state = "error"
            error("error: unknown state: '%s' reached in %s" % (state, line))

def textLineToContentLine(text):
    return ContentLine(*parseLine(text), **{'encoded':True})

def dquoteEscape(param):
    """Return param, or "param" if ',' or ';' or ':' is in param."""
    if param.find('"') >= 0:
        raise VObjectError("Double quotes aren't allowed in paramater values.")
    for char in ',;:':
        if param.find(char) >= 0:
            return '"'+ param + '"'
    return param

def defaultSerialize(obj, buf, lineLength):
    """Encode and fold obj and its children, write to buf or return a string."""

    def foldOneLine(input):
        if type(input) in (str, unicode): input = StringIO.StringIO(input)
        input.seek(0)
        outbuf.write(input.read(lineLength) + CRLF)
        brokenline = input.read(lineLength - 1)
        while brokenline:
            outbuf.write(' ' + brokenline + CRLF)
            brokenline = input.read(lineLength - 1)

    if buf: outbuf = buf
    else: outbuf=StringIO.StringIO()

    if isinstance(obj, Component):
        if obj.useBegin: foldOneLine(u"BEGIN:" + obj.name)
        for child in obj.getSortedChildren():
            #validate is recursive, we only need to validate once
            child.serialize(outbuf, lineLength, validate=False)
        if obj.useBegin: foldOneLine(u"END:" + obj.name)

    elif isinstance(obj, ContentLine):
        startedEncoded = obj.encoded
        #TODO: X- lines should be considered TEXT, and should be encoded as such
        if obj.behavior and startedEncoded: obj.behavior.encode(obj)
        s=StringIO.StringIO() #unfolded buffer
        s.write(obj.name.upper())
        for key, paramvals in obj.params.iteritems():
            s.write(';' + key + '=' + ','.join(map(dquoteEscape, paramvals)))
        s.write(':' + obj.value)
        if obj.behavior and startedEncoded: obj.behavior.decode(obj)
        foldOneLine(s)
        
    if not buf:
        return outbuf.getvalue()


testVCalendar="""
BEGIN:VCALENDAR
BEGIN:VEVENT
SUMMARY;blah=hi!:Bastille Day Party
END:VEVENT
END:VCALENDAR"""

def readComponents(stream, validate=False, transform=True):
    """Generate one Component at a time from a stream.

    >>> import StringIO
    >>> f = StringIO.StringIO(testVCalendar)
    >>> cal=readComponents(f).next()
    >>> cal
    <VCALENDAR| [<VEVENT| [<SUMMARY{u'BLAH': [u'hi!']}Bastille Day Party>]>]>
    >>> cal.vevent[0].summary[0]
    <SUMMARY{u'BLAH': [u'hi!']}Bastille Day Party>
    
    """
    #stack and helpers for it
    stack = []
    def topVobj():
        if len(stack) == 0: return None
        else: return stack[-1]
    def topName():
        if len(stack) == 0: return None
        else: return stack[-1].name
    def addToTopOfStack(item):
        if topVobj(): topVobj().add(item)
        else:
            push(Component())
            topVobj().add(item)
    def push(vObj): stack.append(vObj)
    def pop(): return stack.pop()


    for line in getLogicalLines(stream):
        vline = textLineToContentLine(line)
        if vline.name == "VERSION": topVobj().setBehaviorFromVersionLine(vline)

        if   vline.name == "BEGIN":   push(Component(vline.value))
        elif vline.name == "PROFILE":
            if not topVobj(): push(Component())
            topVobj().setProfile(vline.value)
        elif vline.name == "END":
            if len(stack) == 0:
                err = "Attempted to end the %s component, \
                       but it was never opened" % vline.value
                raise VObjectError(err)
            if vline.value == topName():#START matches END
                if len(stack) == 1:
                    component=pop()
                    if validate: component.validate(raiseException=True)
                    if transform: component.transformChildrenToNative()
                    yield component #EXIT POINT
                else: addToTopOfStack(pop())
            else: raise VObjectError("%s component wasn't closed" % topName())
        else: addToTopOfStack(vline) #not a START or END line
    if topVobj():
        if topVobj().name is None:
            logger.warning("Top level component was never named")
        elif topVobj().useBegin:
            raise VObjectError("Component %s was never closed" % (topName()))
        yield pop()


#------------------- Testing and running functions -----------------------------
if __name__ == '__main__':
    import tests
    tests._test()

#------------------------- Removed but still useful ----------------------------

def serializeQPLine(self, buf=None, lineLength=75, rfc2425=True):
    """Encode and fold this line, write to buf or return."""
    if not rfc2425:
        lineLength -= 1 # The '=' character needs taking into account.
    s=StringIO.StringIO()
    s.write( str(self.name) )
    if len(self.params):
        s.write( ";" )
        s.write( ';'.join([str(x) + '=' + str(z) for x,y in self.params.items() for z in y]) )
    if len(self.singletonparams):
        if rfc2425:
            raise VObjectError("Unclaimed and unnamed parameters while trying to write RFC2425 format.")
        s.write( ';' + ';'.join(self.singletonparams) )
    qp = False
    sv = str(self.value)
    if sv.find('\n')!=-1:
        if rfc2425:
            sv = sv.replace('\n','\\n')
        else:
            qp = True
    if not rfc2425:
        if (s.tell() + len(sv)) >= lineLength:
            qp = True
    if qp:
        s.write( ';ENCODING=QUOTED-PRINTABLE:' )
        sv = sv.encode('quoted-printable').replace('=\n','').replace('\r','' ).replace('\n','=%02X' % ord('\n') )
        written = False
        while len(sv):
            if written:
                s.write( '=\r\n' )
            s.write( sv[:lineLength] )
            sv = sv[lineLength:]
            written = True
    else:
        s.write( ':' )
        s.write( sv )
    s.write('\r\n')
    s.seek(0)
    s2 = buf
    if s2 is None:
        s2 = StringIO.StringIO()
    if rfc2425:
        tmp = s.getvalue()
        wr = False
        while len(tmp):
            if wr:
                s.write( '\r\n ' )
            s2.write( tmp[:lineLength] )
            tmp = tmp[lineLength:]
            wr = True
    else:
        s2.write(s.getvalue())
    if buf is None:
        return s2.getvalue()
